#include "pest.h"
#include "spec.h"
#include "packet.h"
#include "chain.h"
#include "crypto.h"
#include "peer.h"
#include "serial.h"
#include "string.h"
#include "utf8.h"
#include "time.h"
#include "log.h"
#include "state.h"
#include <assert.h>

bool prod_send(peer_t *p,uint16_t flag){
  uint8_t     t[PROD_SIZE];
  uint16_t    port = p->addr.port;
  uint32_t    ip   = byteswap32(p->addr.ip);
  bool        ret;
  char const *bann;

  if(!peer_cansend(p,0)) return 1;

  assert(flag == PROD_FLAG_REQANS ||
         flag == PROD_FLAG_ANSWER);

  write_u16(t+PROD_FLAG_OFF,  &flag);
  write_u16(t+PROD_PORT_OFF,  &port);
  write_u32(t+PROD_IP_OFF,    &ip);

  hash_copy(t+PROD_BCSC_OFF,  self);
  hash_copy(t+PROD_BCNC_OFF,  net);
  hash_copy(t+PROD_DMSC_OFF,  p->self);

  bann = *banner ? banner : PROG_FULLERNAME;
  strncpy((char *)t+PROD_BANN_OFF,bann,PROD_BANN_SIZE);

  ret = send_text(CMD_PROD,mono_sec(),t,sizeof(t),p,0);
  BUF_ZERO(t);
  return ret;
}

static void getdata_maybe(peer_t       *p,
                          peer_t       *chan,
                          ht_t         *ht,
                          mess_t       *ms,
                          hash_t const  hash){

  if(ht_lookup(ht,hash))                        return;
  if(mess_lookup(ms,hash,0) != LOOKUP_NOTFOUND) return;
  getdata_push_cond(p,chan,hash);
}

void prod_handle(peer_t *p,uint8_t const *t){
  uint16_t  flag;
  uint16_t  port;
  uint32_t  ip;
  char      bann[BANNER_BUF];

  read_u16(t+PROD_FLAG_OFF,   &flag);
  read_u16(t+PROD_PORT_OFF,   &port);
  read_u32(t+PROD_IP_OFF,     &ip);

  strzzcpy(bann,
           (char const *)(t+PROD_BANN_OFF),
           sizeof(bann),
           PROD_BANN_SIZE);

  if(flag != PROD_FLAG_REQANS &&
     flag != PROD_FLAG_ANSWER){
    log_err("invalid prod flag %u received from peer '%s'",
            flag,p->handle[0]);
    return;
  }

  if(!utf8_valid(bann)){
    log_err("banner from peer '%s' contains invalid UTF-8 sequence",
            p->handle[0]);
    return;
  }

  acaddr.port = port;
  acaddr.ip   = byteswap32(ip);

  if(flag == PROD_FLAG_REQANS){
    prod_send(p,PROD_FLAG_ANSWER);
  }

  getdata_maybe(p,0,&p->bhead,&mess,t+PROD_BCSC_OFF);
  getdata_maybe(p,p,&p->dhead,&p->mess,t+PROD_DMSC_OFF);

  if(strcmp(bann,p->banner) != 0){
    peer_banner_set(p,bann);
  }
}

void prod_periodic(uint64_t time){
  if(time < prod_last+prodint) return;
  prod_last = time;

  for(size_t x = 0;x < peer_cnt;x++){
    peer_t *p = peer+x;
    prod_send(p,PROD_FLAG_REQANS);
  }
}

